from scadgen import scadobj

ScadContext = scadobj.ScadContext
ScadModule = scadobj.ScadModule
MissingScadContextException = scadobj.MissingScadContextException

circle = scadobj.define_ScadObj('circle')
square = scadobj.define_ScadObj('square')
polygon = scadobj.define_ScadObj('polygon')
text = scadobj.define_ScadObj('text')
projection = scadobj.define_ScadOperation('projection')

sphere = scadobj.define_ScadObj('sphere')
cube = scadobj.define_ScadObj('cube')
cylinder = scadobj.define_ScadObj('cylinder')
polyhedron = scadobj.define_ScadObj('polyhedron')
linear_extrude = scadobj.define_ScadOperation('linear_extrude')
rotate_extrude = scadobj.define_ScadOperation('rotate_extrude')

translate = scadobj.define_ScadOperation('translate')
rotate = scadobj.define_ScadOperation('rotate')
scale = scadobj.define_ScadOperation('scale')
resize = scadobj.define_ScadOperation('resize')
mirror = scadobj.define_ScadOperation('mirror')
multmatrix = scadobj.define_ScadOperation('multmatrix')
color = scadobj.define_ScadOperation('color')
offset = scadobj.define_ScadOperation('offset')
hull = scadobj.define_ScadOperation('hull')
minkowski = scadobj.define_ScadOperation('minkowski')

union = scadobj.define_ScadOperation('union')
difference = scadobj.define_ScadOperation('difference')
intersection = scadobj.define_ScadOperation('intersection')

print_tree = scadobj.print_tree



from __future__ import annotations
import contextlib
import abc
from typing import Generator, Optional

# The top level global context object.
_GENSCAD_GLOBAL_CONTEXT = None


def arg_value_to_str(v):
    '''Convert from Python to OpenSCAD representation.'''
    if v == True:
        return 'true'
    elif v == False:
        return 'false'
    elif isinstance(v, str):
        return '"' + str(v) + '"'
    else:
        return str(v)


def format_openscad_decl(keyword, depth, args, kwargs):
    '''Format an OpenSCAD keyword and args.'''
    str_args = [arg_value_to_str(a) for a in args]
    str_kwargs = [f'{k}={arg_value_to_str(v)}' for k, v in kwargs.items()]
    args = ', '.join(str_args+str_kwargs)
    return '  '*(depth-1) + f'{keyword}({args})'


def print_tree(o):
    '''Print the object tree rooted at a context object.'''
    print('  '*o.depth() + str(repr(o)))
    try:
        for module in o.modules:
            print_tree(module)
    except AttributeError:
        pass
    try:
        for child in o.objs:
            print_tree(child)
    except AttributeError:
        # ScadObj does not have a .objs member.
        pass


class ScadEntity(abc.ABC):
    '''Parent class for any object that generates an OpenSCAD statement.'''
    parent_entity: Optional[ScadEntity] = None

    def gen(self) -> str:
        lines = list(self._generate())
        # lines = [l + ' // depth: ' + str(self.depth()) for l in lines]
        return '\n'.join(lines)

    @abc.abstractmethod
    def _generate() -> Generator[str, None, None]:
        raise NotImplementedError()

    def depth(self):
        if self.parent_entity is None:
            return 0
        else:
            return 1 + self.parent_entity.depth()


class ScadContext(contextlib.AbstractContextManager, ScadEntity):
    '''Context manager.'''
    _NAME = '???'

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.objs = []
        self.modules = []

    def __enter__(self):
        global _GENSCAD_GLOBAL_CONTEXT
        self.parent_entity = _GENSCAD_GLOBAL_CONTEXT
        if self.parent_entity is not None:
            self.parent_entity.add_obj(self)
        _GENSCAD_GLOBAL_CONTEXT = self
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        global _GENSCAD_GLOBAL_CONTEXT
        _GENSCAD_GLOBAL_CONTEXT = self.parent_entity
        return False

    def __repr__(self) -> str:
        return(f'ScadContext_{self._NAME}(args={self.args}, kwargs={self.kwargs})')

    def _generate(self) -> Generator[str, None, None]:
        for m in self.modules:
            yield m.gen()
        for o in self.objs:
            yield o.gen()

    def add_obj(self, obj: ScadEntity):
        self.objs.append(obj)

    def add_module(self, module: ScadModule):
        if self.parent_entity:
            self.parent_entity.add_module(module)
        else:
            if module._NAME not in [m._NAME for m in self.modules]:
                self.modules.append(module)
                module.parent_entity = self


class ScadModule(ScadContext):
    # TODO: Module parameters.
    def __init__(self, name, *args, **kwargs):
        super(ScadModule, self).__init__(*args, **kwargs)
        self._NAME = name
        self.parent_entity = _GENSCAD_GLOBAL_CONTEXT
        self.obj_cls = define_ScadObj(self._NAME)

    def __repr__(self) -> str:
        return(f'ScadModule_{self._NAME}(args={self.args}, kwargs={self.kwargs})')

    def __call__(self, *args, **kwargs):
        # Calling a module inserts a ScadObj with that name.
        self.obj_cls(*args, **kwargs)
        # We also need to make sure the module and its dependencies are registered.
        _GENSCAD_GLOBAL_CONTEXT.add_module(self)
        for m in self.modules:
            _GENSCAD_GLOBAL_CONTEXT.add_module(m)

    def _generate(self) -> Generator[str, None, None]:
        yield '  ' * (self.depth()-1) + f'module {self._NAME}() {{'
        for o in self.objs:
            yield o.gen()
        yield '  ' * (self.depth()-1) + '}'


class ScadOperation(ScadContext):
    '''An OpenSCAD operation that acts on objects in its brace block.'''

    def __init__(self, *args, **kwargs):
        global _GENSCAD_GLOBAL_CONTEXT
        if not _GENSCAD_GLOBAL_CONTEXT:
            raise MissingScadContextException('No enclosing ScadContext.')
        super(ScadOperation, self).__init__(*args, **kwargs)
        self.args = args
        self.kwargs = kwargs
        self.is_nested = False

    def __repr__(self) -> str:
        return(f'ScadOperation_{self._NAME}(args={self.args}, kwargs={self.kwargs})')

    def _generate(self) -> Generator[str, None, None]:
        yield format_openscad_decl(self._NAME, self.depth(), self.args, self.kwargs) + ' {'
        for o in self.objs:
            yield o.gen()
        yield '  ' * (self.depth()-1) + '}'

    def __add__(self, o: ScadOperation):
        return ScadCompositeOperation(self, o)


class ScadCompositeOperation(ScadOperation):
    # Enables composite operations using '+' shorthand.
    def __init__(self, a, b, *args, **kwargs):
        super(ScadCompositeOperation, self).__init__(*args, **kwargs)
        self.a = a
        self.b = b
        self.a.objs = [self.b]
        self.b.parent_entity = self.a
        self.last = self.b

    def __repr__(self) -> str:
        return(f'ScadCompositeOperation({self.a}, {self.b})(args={self.args}, kwargs={self.kwargs})')

    def __add__(self, o: ScadOperation):
        self.last.add_obj(o)
        o.parent_entity = self.last
        self.last = o
        return self

    def add_obj(self, obj: ScadEntity):
        return self.b.add_obj(obj)

    def __enter__(self):
        global _GENSCAD_GLOBAL_CONTEXT
        self.a.parent_entity = _GENSCAD_GLOBAL_CONTEXT
        if self.a.parent_entity is not None:
            self.a.parent_entity.add_obj(self.a)
        _GENSCAD_GLOBAL_CONTEXT = self.last
        return self.last

    def __exit__(self, exc_type, exc_value, traceback):
        global _GENSCAD_GLOBAL_CONTEXT
        _GENSCAD_GLOBAL_CONTEXT = self.a.parent_entity
        return False


class ScadObj(ScadEntity):
    '''An OpenSCAD object, terminated by a semicolon (no brace block).'''
    _NAME = '???'

    def __init__(self, *args, **kwargs):
        global _GENSCAD_GLOBAL_CONTEXT
        if not _GENSCAD_GLOBAL_CONTEXT:
            raise MissingScadContextException('No enclosing ScadContext.')
        self.args = args
        self.kwargs = kwargs
        _GENSCAD_GLOBAL_CONTEXT.add_obj(self)
        self.parent_entity = _GENSCAD_GLOBAL_CONTEXT

    def __repr__(self) -> str:
        return(f'ScadObj_{self._NAME}(args={self.args}, kwargs={self.kwargs})')

    def _generate(self) -> Generator[str, None, None]:
        yield format_openscad_decl(self._NAME, self.depth(), self.args, self.kwargs) + ';'


def define_ScadOperation(name):
    '''Helper function to define a named ScadOperation.'''
    class _ScadOperation(ScadOperation):
        _NAME = name
    return _ScadOperation


def define_ScadObj(name):
    '''Helper function to define a named ScadObj.'''
    class _ScadObj(ScadObj):
        _NAME = name
    return _ScadObj


class MissingScadContextException(Exception):
    pass



import unittest
import textwrap
import scadgen as s


class TestModule(unittest.TestCase):

    def test_basic_module(self):
        '''Basic module usage.'''
        with s.ScadModule(name='cylcube') as cylcube:
            # A module comprising a cylinder with a cube stacked on top of it
            s.cylinder(20, r=20)
            with s.translate([0, 0, 25]):
                s.cube(size=[10, 10, 10], center=True)
        with s.ScadContext() as model:
            cylcube()
            with s.translate([100, 0, 0]):
                cylcube()
        output = model.gen()
        self.assertEqual(output, textwrap.dedent('''\
            module cylcube() {
              cylinder(20, r=20);
              translate([0, 0, 25]) {
                cube(size=[10, 10, 10], center=true);
              }
            }
            cylcube();
            translate([100, 0, 0]) {
              cylcube();
            }'''))

    def test_dependent_module(self):
        '''A module that depends on another module.'''
        with s.ScadModule(name='cylcube') as cylcube:
            s.cylinder(20, r=20)
            with s.translate([0, 0, 25]):
                s.cube(size=[10, 10, 10], center=True)
        with s.ScadModule(name='doublecylcube') as doublecylcube:
            cylcube()
            with s.translate([0, 0, 30]):
                cylcube()
        with s.ScadContext() as model:
            doublecylcube()
            with s.translate([100, 0, 0]):
                doublecylcube()
        output = model.gen()
        # print('_____ model tree _____ ')
        # s.print_tree(model)
        # print('_____ model output _____ ')
        # print(output)
        self.assertEqual(output, textwrap.dedent('''\
            module doublecylcube() {
              cylcube();
              translate([0, 0, 30]) {
                cylcube();
              }
            }
            module cylcube() {
              cylinder(20, r=20);
              translate([0, 0, 25]) {
                cube(size=[10, 10, 10], center=true);
              }
            }
            doublecylcube();
            translate([100, 0, 0]) {
              doublecylcube();
            }'''))


class TestContext(unittest.TestCase):

    def test_context_basic(self):
        '''Test that we can enter/exit a context.'''
        with s.ScadContext():
            pass

    def test_object_no_context(self):
        with self.assertRaises(s.MissingScadContextException):
            s.cube(size=[10, 10, 10], center=False)

    def test_operation_no_context(self):
        with self.assertRaises(s.MissingScadContextException):
            with s.translate([10, 10, 10]):
                pass

    def test_basic_obj(self):
        '''Basic usage.'''
        with s.ScadContext() as model:
            s.cube(size=[10, 10, 10], center=False)
            with s.translate([100, 100, 100]):
                s.cube(size=[20, 20, 20], center=False)
        output = model.gen()
        self.assertEqual(output, textwrap.dedent('''\
            cube(size=[10, 10, 10], center=false);
            translate([100, 100, 100]) {
              cube(size=[20, 20, 20], center=false);
            }'''))

    def test_basic2_obj(self):
        '''Slightly more complex basic usage.'''
        with s.ScadContext() as model:
            s.cube(size=[25, 25, 25], center=False)
            for x in [100, 200, 300]:
                with s.translate([x, 0, 0]):
                    s.cube(size=[20, 20, 20], center=False)
                    s.cylinder(100, r=5, center=True)
        output = model.gen()
        self.assertEqual(output, textwrap.dedent('''\
            cube(size=[25, 25, 25], center=false);
            translate([100, 0, 0]) {
              cube(size=[20, 20, 20], center=false);
              cylinder(100, r=5, center=true);
            }
            translate([200, 0, 0]) {
              cube(size=[20, 20, 20], center=false);
              cylinder(100, r=5, center=true);
            }
            translate([300, 0, 0]) {
              cube(size=[20, 20, 20], center=false);
              cylinder(100, r=5, center=true);
            }'''))

    def test_function_obj(self):
        '''Nested translates and a helper function.'''
        def helper_fn():
            s.cylinder(10, r=16, center=True)
            with s.translate([0, 0, -10]):
                s.cylinder(30, r=8, center=True)

        with s.ScadContext() as model:
            for x in [100, 200, 300]:
                with s.translate([x, 0, 0]):
                    helper_fn()
        output = model.gen()
        self.assertEqual(output, textwrap.dedent('''\
            translate([100, 0, 0]) {
              cylinder(10, r=16, center=true);
              translate([0, 0, -10]) {
                cylinder(30, r=8, center=true);
              }
            }
            translate([200, 0, 0]) {
              cylinder(10, r=16, center=true);
              translate([0, 0, -10]) {
                cylinder(30, r=8, center=true);
              }
            }
            translate([300, 0, 0]) {
              cylinder(10, r=16, center=true);
              translate([0, 0, -10]) {
                cylinder(30, r=8, center=true);
              }
            }'''))


class TestValueconversion(unittest.TestCase):

    def test_value_conversion(self):
        '''Param value conversion from Python to OpenSCAD'''
        with s.ScadContext() as model:
            s.cube(size=[10, 10, 10], center=False,
                   some_str_arg='foo', some_float_arg=0.333)
        output = model.gen()
        self.assertEqual(output, textwrap.dedent('''\
            cube(size=[10, 10, 10], center=false, some_str_arg="foo", some_float_arg=0.333);'''))

    def test_value_conversion_args(self):
        '''Param string conversion.'''
        with s.ScadContext() as model:
            with s.color("blue"):
                s.cube()
        output = model.gen()
        self.assertEqual(output, textwrap.dedent('''\
            color("blue") {
              cube();
            }'''))


class TestCompositeOperation(unittest.TestCase):
    def test_context_add(self):
        '''Ability to use the + operator with ScadContexts as a shorthand.'''
        # The basic syntax, using nested contexts.
        with s.ScadContext() as model1:
            with s.translate([1, 1, 1]):
                with s.translate([2, 2, 2]):
                    with s.translate([3, 3, 3]):
                        s.cube()
        # Adding operations as a shorthand.
        with s.ScadContext() as model2:
            with s.translate([1, 1, 1]) + s.translate([2, 2, 2]) + s.translate([3, 3, 3]):
                s.cube()
        self.assertEqual(model1.gen(), model2.gen())



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['scadgen']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'scadgen',
    'version': '0.0.2',
    'description': 'A Python library for generating OpenSCAD 3D models.',
    'long_description': '# scadgen\n\nA Python library for generating OpenSCAD 3D models.\n\n',
    'author': 'TC Teo',
    'author_email': 'scadgen-pypi-011ecc47@tsechin.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/tcteo/scadgen',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
